// ==== 6 digits Frequency Counter ====
// TMR1 counts from 32768 to 65536 and the crystal is 32768Hz. This give us exactly 1 second time interval.
// The counter can be callibrated with setting different starting value of TMR1
// TMR0 counts the input signal. Each overflow of TMR0 is counted in f2.
// 8-bit TMR2 is used for refreshing the LED display. It uses the internal oscillator with 1MHz frequency.
// the postscaller of TMR2 is set to 1:13. TMR2 sets overflow flag at every 13 x 256 impulses.
// This gives us 50Hz refreshing frequency of every digit.
// RA0, RA1, RA2, RA3 and RA7 are used for digit selection. 
// Setting low one of these enables the corresponding digit.
// RB0..RB5 and RA6 flashes the segments of the digit.

#include <htc.h>
#define _XTAL_FREQ  4000000
#define byte        unsigned char
#define u_int       unsigned int
#define u_long      unsigned long
#define OFF         10

__CONFIG(FOSC_INTOSCIO & MCLRE_ON & WDTE_OFF & CPD_OFF & CP_OFF & LVP_OFF & BOREN_ON & PWRTE_ON);

//  Digit selection
const byte digits[6] = {
  0b10111111,   // digit 0: all high
  0b10110111,   // digit 1: RA3 - low
  0b10111011,   // digit 2: RA2 - low
  0b10111101,   // digit 3: RA1 - low
  0b10111110,   // digit 4: RA0 - low
  0b00111111,   // digit 5: RA7 - low
};

//     RA6 RB5 RB4 RB3 RB2 RB1 RB0
//     f   a   e   g   d   c   b
//  0: 1   1   1   0   1   1   1   0x77
//  1: 0   0   0   0   0   1   1   0x03
//  2: 0   1   1   1   1   0   1   0x3D
//  3: 0   1   0   1   1   1   1   0x2F
//  4: 1   0   0   1   0   1   1   0x4B
//  5: 1   1   0   1   1   1   0   0x6E
//  6: 1   1   1   1   1   1   0   0x7E
//  7: 0   1   0   0   0   1   1   0x23
//  8: 1   1   1   1   1   1   1   0x7F
//  9: 1   1   0   1   1   1   1   0x6F
// BL: 0   0   0   0   0   0   0   0x00

const byte segments[11] = {
  0b110111,   // 0
  0b000011,   // 1
  0b111101,   // 2
  0b101111,   // 3
  0b001011,   // 4
  0b101110,   // 5
  0b111110,   // 6
  0b100011,   // 7
  0b111111,   // 8
  0b101111,   // 9
  0b000000    // OFF
};

// segment F is switched by pin RA6
const byte segmentF[11] = {
  0b01000000,  // 0
  0b00000000,  // 1
  0b00000000,  // 2
  0b00000000,  // 3
  0b01000000,  // 4
  0b01000000,  // 5
  0b01000000,  // 6
  0b00000000,  // 7
  0b01000000,  // 8
  0b01000000,  // 9
  0b00000000   // OFF
};

byte n, k, d[6];			// global variables

void display() {			// displays n-th digit
  byte dig;
  dig = d[n];
  PORTB = 0;
  PORTA = 0;
  PORTA = digits[n] | segmentF[dig];
  PORTB = segments[dig];
  if (n > 0) n--;
  else n = k;
  TMR2IF = 0;
}

void main(){
  u_long frequency;
  u_int f2;
  byte i;
  OSCF = 1;                 // set the internal oscillator at 4MHz
  CMCON = 0b00000111;       // comparator off
  T1CON = 0b00001110;       // TMR1 prescale 1:1, osc = on
  TRISB = 0b11000000;       // RB0..RB5 - segments A,B,C,D,E,G
  TRISA = 0b00110000;       // RA0..RA3, RA7 - outputs for digit selection, RA6 - segment F
  OPTION_REG = 0b10101000;  // TMR0
  T2CON = 0b01100100;     	// Timer2 on, postscaller 1:13, prescaller 1:1
  PR2 = 0xFF;
  d[0] = 0;
  d[1] = OFF;
  d[2] = OFF;
  d[3] = OFF;
  d[4] = OFF;
  d[5] = OFF;
  n = 0;
  k = 0;

  while (1) {               // main loop
    f2 = 0;
    // with different start values of TMR1L and TMR1H we can fine tune the time interval if needed. 
	TMR1L = 0;
    TMR1H = 128;            // start counting from 32768
    TMR1IF = 0;				
    T0IF = 0;				// clear the interrupt flag of Timer0
    TMR1ON = 1;             // start Timer1
    TMR0 = 0;               // start counting input signal (Timer0)
    do {					// loop for 1 second
	  if (T0IF) {
        T0IF = 0;           // clear T0IF bit
        f2++;
      }
      if (TMR2IF) display();		// lighting the next digit
    } while (!TMR1IF);
    
    frequency = (u_long)f2*256 + (u_long)TMR0;  // calculating current frequency
    
	TMR1ON = 0;             //stop Timer1

    // Separating digits
	// While calculating, the program continue to flash digits, to reduce fickering at every second
    i = 0;
	do {
      d[i] = frequency % 10;
	  if (TMR2IF) display();		// lighting the next digit
	  frequency = frequency / 10;
	  if (TMR2IF) display();		// lighting the n-th digit
	} while (++i < 6);

    // switch off leading zeroes.
	k = 5;
    if (!d[5]) {
      PORTA = digits[5] | segmentF[OFF];
      PORTB = segments[OFF];
	  k = 4; 
      if (!d[4]) {
        PORTA = digits[4] | segmentF[OFF];
        PORTB = segments[OFF];
		k = 3;
        if (!d[3]) {
          PORTA = digits[3] | segmentF[OFF];
      	  PORTB = segments[OFF];
		  k = 2;
          if (!d[2]) { 
			PORTA = digits[2] | segmentF[OFF];
      		PORTB = segments[OFF];
		    k = 1;
			if (!d[1]) {
			  PORTA = digits[2] | segmentF[OFF];
      		  PORTB = segments[OFF];
		      k = 0;
			}
		  }
        }
      }
    }
  }
}